#include <J3ML/LinearAlgebra/Transform2D.hpp>

namespace J3ML::LinearAlgebra {

    const Transform2D Transform2D::Identity = Transform2D({0, 0}, {1, 1}, {0,0}, {0,0}, 0);
    const Transform2D Transform2D::FlipX    = Transform2D({0, 0}, {-1, 1}, {0,0}, {0,0}, 0);
    const Transform2D Transform2D::FlipY    = Transform2D({0, 0}, {1, -1}, {0,0}, {0,0}, 0);


    Vector2 Transform2D::Transform(const Vector2 &input) const {
        return transformation.Transform(input);
    }

    Transform2D::Transform2D(const Matrix3x3 &transform) : transformation(transform) { }

    Transform2D::Transform2D(const Vector2& pos, const Vector2& scale, const Vector2& origin, const Vector2& skew, float rotation) {
        transformation = Matrix3x3(pos.x, pos.y, rotation, scale.x, scale.y, origin.x, origin.y, skew.x, skew.y);
    }

    Transform2D Transform2D::Translate(float x, float y) const {
        auto copy = Matrix3x3(transformation);
        copy.SetAt(0, 0, transformation.At(0, 0) + x);
        copy.SetAt(0, 1, transformation.At(0, 1) + y);
        return Transform2D(copy);
    }

    Transform2D Transform2D::Translate(const LinearAlgebra::Vector2 &input) const {
        return Translate(input.x, input.y);
    }

    Transform2D::Transform2D() {
        transformation = Matrix3x3::Identity;
    }

    Transform2D Transform2D::FromRotation(float radians) {
        float c = Math::Cos(radians);
        float s = Math::Sin(radians);
        return Transform2D(Matrix3x3(
                c, s, 0.f,
                -s, c, 0.f,
                0.f, 0.f, 1.f));
    }

    Transform2D Transform2D::FromScale(const Vector2 &scale) {
        return FromScale(scale.x, scale.y);
    }

    Transform2D Transform2D::FromScale(float sx, float sy) {
        Transform2D s;
        s.transformation[0][0] = sx;
        s.transformation[1][1] = sy;
        return s;
    }

    Transform2D Transform2D::FromTranslation(const Vector2 &translation) {
        return FromTranslation(translation.x, translation.y);
    }

    Transform2D Transform2D::FromTranslation(float tx, float ty) {
        return Transform2D(Matrix3x3(
                1.f, 0.f, 0.f,
                0.f, 1.f, 0.f,
                tx,  ty,  1.f));
    }

    Vector2 Transform2D::GetScale() const {
        return {
                Math::Sqrt(At(0,0) * At(0,0) + At(0,1) * At(0,1)),
                Math::Sqrt(At(1,0) * At(1,0) + At(1,1) * At(1,1))
        };
    }

    float Transform2D::GetRotation() const {
        return Math::Atan2(At(1, 0), At(0, 0));
    }

    Vector2 Transform2D::GetTranslation() const {
        return {At(2,0), At(2, 1)};
    }

    float &Transform2D::At(int row, int col) { return transformation.At(row, col); }

    float Transform2D::At(int row, int col) const { return transformation.At(row, col); }

    float Transform2D::Determinant() const { return transformation.Determinant(); }

    /*
    Vector2 Transform2D::Transform(const Vector2 &point) const {
        Vector2 result;
        result.x = At(0,0) * point.x + At(0,1) * point.y + At(0,2);
        result.y = At(1,0) * point.x + At(1,1) * point.y + At(1,2);
        return result;
    }
     */

    Vector2 Transform2D::ForwardVector() const { return {At(0,0), At(1,0)}; }

    Vector2 Transform2D::UpVector() const { return {At(0,1), At(1,1)}; }
}